/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.activiti.doclet;

import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.nio.charset.Charset;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.Name;
import javax.lang.model.element.VariableElement;
import javax.lang.model.type.TypeMirror;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.sun.source.doctree.BlockTagTree;
import com.sun.source.doctree.DocCommentTree;
import com.sun.source.doctree.DocTree;
import com.sun.source.doctree.InlineTagTree;
import com.sun.source.util.DocTrees;

import freemarker.template.Configuration;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import freemarker.template.TemplateExceptionHandler;
import jdk.javadoc.doclet.DocletEnvironment;

/**
 * @author brian@inteligr8.com
 */
class MarkdownWriter {

	private final Logger logger = LoggerFactory.getLogger(this.getClass());
	private final Pattern tagPattern = Pattern.compile("@[^ ]+ ?(.*)$");
	private final Pattern namedTagPattern = Pattern.compile("@[^ ]+ ([^ ]+) ?(.*)$");
	private final DocletEnvironment docenv;
	private final File outputDirectory;
	private final Configuration fmconfig;
	private String title = "API Documentation";
	private String apiName;
	private String flavor;
	
	public MarkdownWriter(DocletEnvironment docenv, String outputDirectory) throws IOException {
		this.docenv = docenv;
		this.outputDirectory = (outputDirectory == null || outputDirectory.trim().length() == 0) ? new File(".") : new File(outputDirectory);
		this.fmconfig = this.getFreemarkerConfiguration();
	}
	
	public MarkdownWriter(DocletEnvironment docenv, File outputDirectory) throws IOException {
		this.docenv = docenv;
		this.outputDirectory = outputDirectory == null ? new File(".") : outputDirectory;
		this.fmconfig = this.getFreemarkerConfiguration();
	}
	
	private Configuration getFreemarkerConfiguration() throws IOException {
		Configuration fmconfig = new Configuration(Configuration.VERSION_2_3_31);
		fmconfig.setDefaultEncoding("utf-8");
		fmconfig.setTemplateExceptionHandler(TemplateExceptionHandler.RETHROW_HANDLER);
		fmconfig.setLogTemplateExceptions(false);
		fmconfig.setWrapUncheckedExceptions(true);
		fmconfig.setFallbackOnNullLoopVariable(false);
		fmconfig.setClassLoaderForTemplateLoading(this.getClass().getClassLoader(), "/templates");
		return fmconfig;
	}
	
	public void setTitle(String title) {
		this.title = title;
	}
	
	public void setApiName(String apiName) {
		this.apiName = apiName;
	}
	
	public void setFlavor(String flavor) {
		this.flavor = flavor;
	}
	
	public void write(List<ActivitiApiBeanDoc> beandocs) throws IOException, TemplateException {
		IndexFreemarkerModel model = this.buildModel(beandocs);
		
		this.outputDirectory.mkdirs();
		this.writeIndexFile(model);
		for (BeanFreemarkerModel beanModel : model.getBeans())
			this.writeBeanFile(beanModel);
	}
	
	private IndexFreemarkerModel buildModel(List<ActivitiApiBeanDoc> beandocs) throws IOException, TemplateException {
		DocTrees docs = this.docenv.getDocTrees();

		this.logger.debug("Building documentation index model");
		
		IndexFreemarkerModel indexModel = new IndexFreemarkerModel();
		System.out.println("title: " + this.title);
		indexModel.setTitle(this.title);
		indexModel.setApiName(this.apiName);

		for (ActivitiApiBeanDoc beandoc : beandocs) {
			DocCommentTree commentTree = docs.getDocCommentTree(beandoc.getClassElement());

			this.logger.debug("Building documentation bean model: {}", beandoc.getBeanId());
	
			BeanFreemarkerModel beanModel = new BeanFreemarkerModel();
			beanModel.setTitle(this.title);
			beanModel.setApiName(this.apiName);
			beanModel.setBeanId(beandoc.getBeanId());
			beanModel.setBookmark(this.formatBookmark(beanModel.getBeanId()));
			if (beandoc.isDelegate())
				beanModel.getDelegateRoles().add("ServiceTask");
			if (beandoc.isExecutionListener())
				beanModel.getDelegateRoles().add("ExecutionListener");
			if (beandoc.isTaskListener())
				beanModel.getDelegateRoles().add("TaskListener");
			if (commentTree != null) {
				this.logger.debug("Found documentation for bean: {}", beandoc.getBeanId());
				
				if (commentTree.getFirstSentence() != null)
					beanModel.setDocFirstSentence(commentTree.getFirstSentence().toString());
				if (commentTree.getBody() != null)
					beanModel.setDocBody(commentTree.getFullBody().toString());
				for (DocTree tag : commentTree.getBlockTags()) {
					String tagName = null;
					if (tag instanceof BlockTagTree) {
						tagName = ((BlockTagTree)tag).getTagName();
					} else if (tag instanceof InlineTagTree) {
						tagName = ((InlineTagTree)tag).getTagName();
					} else {
						this.logger.error("A tag comment is not of an expected type: {}", tag.getClass());
						continue;
					}

					this.logger.trace("Found tag '{}' for bean: {}", tagName, beandoc.getBeanId());
					this.putAdd(beanModel.getTags(), tagName, tag.toString());
				}
			}

			Map<Name, MethodFreemarkerModel> methodOverloads = new LinkedHashMap<>();
			
			for (ExecutableElement methodElement : beandoc.getSortedMethodElements()) {
				DocCommentTree methodCommentTree = docs.getDocCommentTree(methodElement);

				this.logger.debug("Building documentation method model: {}.{}", beandoc.getBeanId(), methodElement);
				
				MethodFreemarkerModel methodModel = methodOverloads.get(methodElement.getSimpleName());
				if (methodModel == null) {
					methodModel = new MethodFreemarkerModel();
					methodModel.setMethodName(methodElement.getSimpleName().toString());
					methodModel.setBookmark(this.formatBookmark(methodModel.getMethodName()));
					
					methodOverloads.put(methodElement.getSimpleName(), methodModel);
					beanModel.getMethods().add(methodModel);
				}
				
				Map<String, String> paramComments = new HashMap<>();
				Map<String, String> throwComments = new HashMap<>();
				
				MethodSignatureFreemarkerModel sigModel = new MethodSignatureFreemarkerModel();
				if (methodCommentTree != null) {
					this.logger.debug("Found documentation for method: {}.{}", beandoc.getBeanId(), methodElement);
					
					if (methodCommentTree.getFirstSentence() != null) {
						if (methodModel.getDocFirstSentence() == null)
							methodModel.setDocFirstSentence(methodCommentTree.getFirstSentence().toString());
						sigModel.setDocFirstSentence(methodCommentTree.getFirstSentence().toString());
					}
					if (methodCommentTree.getBody() != null)
						sigModel.setDocBody(methodCommentTree.getFullBody().toString());
					for (DocTree tag : methodCommentTree.getBlockTags()) {
						String tagName = null;
						if (tag instanceof BlockTagTree) {
							tagName = ((BlockTagTree)tag).getTagName();
						} else if (tag instanceof InlineTagTree) {
							tagName = ((InlineTagTree)tag).getTagName();
						} else {
							this.logger.error("A tag comment is not of an expected type: {}", tag.getClass());
							continue;
						}

						this.logger.trace("Found tag '{}' for method: {}.{}", tagName, beandoc.getBeanId(), methodElement);

						Matcher matcher;
						switch (tagName) {
							case "param":
								matcher = this.namedTagPattern.matcher(tag.toString());
								if (matcher.find())
									paramComments.put(matcher.group(1), matcher.group(2).trim());
								else this.logger.warn("A @param for the {} bean and {} method is improperly formatted", beanModel.getBeanId(), methodElement);
							case "throws":
								matcher = this.namedTagPattern.matcher(tag.toString());
								if (matcher.find())
									throwComments.put(matcher.group(1), matcher.group(2).trim());
								else this.logger.warn("A @throws for the {} bean and {} method is improperly formatted", beanModel.getBeanId(), methodElement);
								break;
							default:
								matcher = this.tagPattern.matcher(tag.toString());
								if (matcher.find())
									this.putAdd(sigModel.getTags(), tagName, matcher.group(1).trim());
								else this.logger.warn("A '@{}' tag for the {} bean and {} method is improperly formatted", tagName, beanModel.getBeanId(), methodElement);
						}
					}
				}
				
				StringBuilder methodSignature = new StringBuilder(methodElement.getSimpleName()).append("(");
				
				for (VariableElement varElement : methodElement.getParameters()) {
					this.logger.trace("Found parameter '{}' for method: {}.{}", varElement.getSimpleName(), beandoc.getBeanId(), methodElement);
					
					ParamFreemarkerModel paramModel = new ParamFreemarkerModel();
					paramModel.setName(varElement.getSimpleName().toString());
					paramModel.setType(varElement.asType().toString());
					paramModel.setComment(paramComments.get(paramModel.getName()));
					
					sigModel.getParams().put(paramModel.getName(), paramModel);
					methodSignature.append(varElement.getSimpleName()).append(": ").append(varElement.asType()).append(", ");
				}
				
				for (TypeMirror thrownType : methodElement.getThrownTypes()) {
					this.logger.trace("Found thrown '{}' for method: {}.{}", thrownType, beandoc.getBeanId(), methodElement);
					
					ThrowFreemarkerModel throwModel = new ThrowFreemarkerModel();
					throwModel.setType(thrownType.toString());
					throwModel.setComment(throwComments.get(throwModel.getType()));
					
					sigModel.getThrowTypes().put(throwModel.getType(), throwModel);
				}

				this.logger.trace("Found return '{}' for method: {}.{}", methodElement.getReturnType(), beandoc.getBeanId(), methodElement);
				if (!"void".equals(methodElement.getReturnType().toString()))
					sigModel.setReturnType(methodElement.getReturnType().toString());
				
				if (methodSignature.charAt(methodSignature.length()-1) != '(')
					methodSignature.delete(methodSignature.length()-2, methodSignature.length());
				methodSignature.append("): ").append(methodElement.getReturnType());
				
				this.logger.debug("Found method signature: {}", methodSignature);
				sigModel.setMethodSignature(methodSignature.toString());
				
				methodModel.getSignatures().add(sigModel);
			}
			
			indexModel.getBeans().add(beanModel);
		}
		
		return indexModel;
	}
	
	private void writeIndexFile(IndexFreemarkerModel model) throws IOException, TemplateException {
		File mdfile = new File(this.outputDirectory, "index.md");
		PrintWriter mdwriter = new PrintWriter(mdfile, Charset.forName("utf-8"));
		try {
			Template template = this.fmconfig.getTemplate("index.md.ftl");
			template.process(model, mdwriter);
		} finally {
			mdwriter.close();
		}
	}
	
	private void writeBeanFile(BeanFreemarkerModel model) throws IOException, TemplateException {
		File mdfile = new File(this.outputDirectory, "bean-" + model.getBeanId() + ".md");
		PrintWriter mdwriter = new PrintWriter(mdfile, Charset.forName("utf-8"));
		try {
			Template template = this.fmconfig.getTemplate("bean.md.ftl");
			template.process(model, mdwriter);
		} finally {
			mdwriter.close();
		}
	}
	
	private String formatBookmark(String title) {
		if (this.flavor == null || title == null)
			return null;
		
		switch (this.flavor.toLowerCase()) {
			case "github":
				return title.toLowerCase().replaceAll(" ", "-");
			case "bitbucket":
				return "markdown-header-" + title.toLowerCase().replaceAll(" ", "-");
			default:
				return null;
		}
	}
	
	private <T> T putAdd(Map<String, List<T>> map, String key, T value) {
		List<T> values = map.get(key);
		if (values == null)
			map.put(key, values = new LinkedList<>());
		values.add(value);
		return value;
	}

}
